<?php

namespace CleantalkSP\Common\Helpers;

use CleantalkSP\Common\Helper;
use CleantalkSP\Variables\Server;
use CleantalkSP\Templates\Singleton;

/**
 * The class contains methods for working with ip
 */
class HelperIP {
    
    use Singleton;
    
    /**
     * @var array Stored IPs
     *            [
     *              [ type ] => IP,
     *              [ type ] => IP,
     *            ]
     */
    private $ips_stored = array();
    
    /**
     * @var array Set of private networks IPv4 and IPv6
     */
    public static $private_networks = array(
        'v4' => array(
            '10.0.0.0/8',
            '100.64.0.0/10',
            '172.16.0.0/12',
            '192.168.0.0/16',
            '127.0.0.1/32',
        ),
        'v6' => array(
            '0:0:0:0:0:0:0:1/128', // localhost
            '0:0:0:0:0:0:a:1/128', // ::ffff:127.0.0.1
        ),
    );
    
    /**
     * Getting arrays of IP (REMOTE_ADDR, X-Forwarded-For, X-Real-Ip, Cf_Connecting_Ip)
     *
     * @param string $ip_type_to_get Type of IP you want to receive
     * @param array  $headers
     *
     * @return string|null
     */
    public static function get( $ip_type_to_get = 'real', $headers = array() )
    {
        // If  return the IP of the current type if it already has been detected
        $ips_stored = self::getInstance()->ips_stored;
        if( ! empty( $ips_stored[ $ip_type_to_get ] ) ){
            return $ips_stored[ $ip_type_to_get ];
        }
        
        $out = null;
        
        switch( $ip_type_to_get ){
            
            // Cloud Flare
            case 'cloud_flare':
                $headers = $headers ?: HelperHTTP::getHeaders();
                if(
                    isset( $headers['Cf-Connecting-Ip'] ) &&
                    ( isset( $headers['Cf-Ray'] ) || isset( $headers['X-Wpe-Request-Id'] ) ) &&
                    ! isset( $headers['X-Gt-Clientip'] )
                ){
                    if( isset( $headers['Cf-Pseudo-Ipv4'], $headers['Cf-Pseudo-Ipv6'] ) ){
                        $source = $headers['Cf-Pseudo-Ipv6'];
                    }else{
                        $source = $headers['Cf-Connecting-Ip'];
                    }
                    $tmp = strpos( $source, ',' ) !== false
                        ? explode( ',', $source )
                        : (array) $source;
                    $ip_version = self::validate( trim( $tmp[0] ) );
                    if( $ip_version ){
                        $out = $ip_version === 'v6' ? self::v6_normalize( trim( $tmp[0] ) ) : trim( $tmp[0] );
                    }
                }
                break;
            
            // GTranslate
            case 'gtranslate':
                $headers = $headers ?: HelperHTTP::getHeaders();
                if( isset( $headers['X-Gt-Clientip'], $headers['X-Gt-Viewer-Ip'] ) ){
                    $ip_version = self::validate( $headers['X-Gt-Viewer-Ip'] );
                    if( $ip_version ){
                        $out = $ip_version === 'v6' ? self::v6_normalize( $headers['X-Gt-Viewer-Ip'] ) : $headers['X-Gt-Viewer-Ip'];
                    }
                }
                break;
            
            // ezoic
            case 'ezoic':
                $headers = $headers ?: HelperHTTP::getHeaders();
                if( isset( $headers['X-Middleton'], $headers['X-Middleton-Ip'] ) ){
                    $ip_version = self::validate( $headers['X-Middleton-Ip'] );
                    if( $ip_version ){
                        $out = $ip_version === 'v6' ? self::v6_normalize( $headers['X-Middleton-Ip'] ) : $headers['X-Middleton-Ip'];
                    }
                }
                break;
            
            // Sucury
            case 'sucury':
                $headers = $headers ?: HelperHTTP::getHeaders();
                if( isset( $headers['X-Sucuri-Clientip'] ) ){
                    $ip_version = self::validate( $headers['X-Sucuri-Clientip'] );
                    if( $ip_version ){
                        $out = $ip_version === 'v6' ? self::v6_normalize( $headers['X-Sucuri-Clientip'] ) : $headers['X-Sucuri-Clientip'];
                    }
                }
                break;
            
            // X-Forwarded-By
            case 'x_forwarded_by':
                $headers = $headers ?: HelperHTTP::getHeaders();
                if( isset( $headers['X-Forwarded-By'], $headers['X-Client-Ip'] ) ){
                    $ip_version = self::validate( $headers['X-Client-Ip'] );
                    if( $ip_version ){
                        $out = $ip_version === 'v6' ? self::v6_normalize( $headers['X-Client-Ip'] ) : $headers['X-Client-Ip'];
                    }
                }
                break;
            
            // Stackpath
            case 'stackpath':
                $headers = $headers ?: HelperHTTP::getHeaders();
                if( isset( $headers['X-Sp-Edge-Host'], $headers['X-Sp-Forwarded-Ip'] ) ){
                    $ip_version = self::validate( $headers['X-Sp-Forwarded-Ip'] );
                    if( $ip_version ){
                        $out = $ip_version === 'v6' ? self::v6_normalize( $headers['X-Sp-Forwarded-Ip'] ) : $headers['X-Sp-Forwarded-Ip'];
                    }
                }
                break;
            
            // Ico-X-Forwarded-For
            case 'ico_x_forwarded_for':
                $headers = $headers ?: HelperHTTP::getHeaders();
                if( isset( $headers['Ico-X-Forwarded-For'], $headers['X-Forwarded-Host'] ) ){
                    $ip_version = self::validate( $headers['Ico-X-Forwarded-For'] );
                    if( $ip_version ){
                        $out = $ip_version === 'v6' ? self::v6_normalize( $headers['Ico-X-Forwarded-For'] ) : $headers['Ico-X-Forwarded-For'];
                    }
                }
                break;
            
            // OVH
            case 'ovh':
                $headers = $headers ?: HelperHTTP::getHeaders();
                if( isset( $headers['X-Cdn-Any-Ip'], $headers['Remote-Ip'] ) ){
                    $ip_version = self::validate( $headers['Remote-Ip'] );
                    if( $ip_version ){
                        $out = $ip_version === 'v6' ? self::v6_normalize( $headers['Remote-Ip'] ) : $headers['Remote-Ip'];
                    }
                }
                break;
            
            // Incapsula proxy
            case 'incapsula':
                $headers = $headers ?: HelperHTTP::getHeaders();
                if( isset( $headers['Incap-Client-Ip'], $headers['X-Forwarded-For'] ) ){
                    $ip_version = self::validate( $headers['Incap-Client-Ip'] );
                    if( $ip_version ){
                        $out = $ip_version === 'v6' ? self::v6_normalize( $headers['Incap-Client-Ip'] ) : $headers['Incap-Client-Ip'];
                    }
                }
                break;
            
            // Remote addr
            case 'remote_addr':
                $ip_version = self::validate( Server::get( 'REMOTE_ADDR' ) );
                if( $ip_version ){
                    $out = $ip_version === 'v6' ? self::v6_normalize( Server::get( 'REMOTE_ADDR' ) ) : Server::get( 'REMOTE_ADDR' );
                }
                break;
            
            // X-Forwarded-For
            case 'x_forwarded_for':
                $headers = $headers ?: HelperHTTP::getHeaders();
                if( isset( $headers['X-Forwarded-For'] ) ){
                    $tmp     = explode( ',', trim( $headers['X-Forwarded-For'] ) );
                    $tmp     = trim( $tmp[0] );
                    $ip_version = self::validate( $tmp );
                    if( $ip_version ){
                        $out = $ip_version === 'v6' ? self::v6_normalize( $tmp ) : $tmp;
                    }
                }
                break;
            
            // X-Real-Ip
            case 'x_real_ip':
                $headers = $headers ?: HelperHTTP::getHeaders();
                if(isset($headers['X-Real-Ip'])){
                    $tmp = explode(",", trim($headers['X-Real-Ip']));
                    $tmp = trim($tmp[0]);
                    $ip_version = self::validate($tmp);
                    if($ip_version){
                        $out = $ip_version === 'v6' ? self::v6_normalize($tmp) : $tmp;
                    }
                }
                break;
            
            // Real
            // Getting real IP from REMOTE_ADDR or Cf_Connecting_Ip if set or from (X-Forwarded-For, X-Real-Ip) if REMOTE_ADDR is local.
            case 'real':
                
                // Detect IP type
                $out = self::get( 'cloud_flare', $headers );
                $out = $out ?: self::get( 'sucury', $headers );
                $out = $out ?: self::get( 'gtranslate', $headers );
                $out = $out ?: self::get( 'ezoic', $headers );
                $out = $out ?: self::get( 'stackpath', $headers );
                $out = $out ?: self::get( 'x_forwarded_by', $headers );
                $out = $out ?: self::get( 'ico_x_forwarded_for', $headers );
                $out = $out ?: self::get( 'ovh', $headers );
                $out = $out ?: self::get( 'incapsula', $headers );
                
                $ip_version = self::validate( $out );
                
                // Is private network
                if(
                    ! $out ||
                    ($out &&
                     (
                         self::is_private_network( $out, $ip_version ) ||
                         self::mask_match(
                             $out,
                             Server::get( 'SERVER_ADDR' ) . '/24',
                             $ip_version
                         )
                     ))
                ){
                    //@todo Remove local IP from x-forwarded-for and x-real-ip
                    $out = $out ?: self::get( 'x_forwarded_for', $headers );
                    $out = $out ?: self::get( 'x_real_ip', $headers );
                }
                
                $out = $out ?: self::get( 'remote_addr', $headers );
                
                break;
            
            default:
                $out = self::get( 'real', $headers );
        }
        
        // Final validating IP
        $out = self::validate( $out )
            ? $out
            : null;
        
        // Store the IP of the current type to skip the work next time
        self::getInstance()->ips_stored[ $ip_type_to_get ] = $out;
        
        return $out;
    }
    
    /**
     * Validating IPv4, IPv6
     *
     * @param mixed $ip
     *
     * @return string|bool
     * 
     */
    public static function validate($ip)
    {
        if(!$ip) return false; // NULL || FALSE || '' || so on...
        if(filter_var($ip, FILTER_VALIDATE_IP, FILTER_FLAG_IPV4) && $ip !== '0.0.0.0') return 'v4';  // IPv4
        if(filter_var($ip, FILTER_VALIDATE_IP, FILTER_FLAG_IPV6) && self::v6_reduce($ip) !== '::') return 'v6';  // IPv6
        return false; // Unknown
    }
    
    /**
     * Expand IPv6
     *
     * Hextet - part of address which could contain 16 bits. Separated by ":" 1:2:3:4:5:6:7:8:
     *
     * @param string $ip
     *
     * @return string IPv6
     */
    public static function v6_normalize($ip)
    {
        if ( ! self::validate($ip) ) return false;
        
        $ip = trim($ip, ' \n\r\t\v"');
        
        // Normalize IPv4 to IPv6
        $ip = preg_match('/^((?>\d{1,3}\.?){4})$/', $ip) ? '0:0:0:0:0:0:' . $ip : $ip;
        $ip = str_replace('::ffff:', '0:0:0:0:0:0:', $ip);
        
        // Searching for ::ffff:xx.xx.xx.xx patterns and turn it to IPv6
        if( preg_match( '/^(.*?)((?>\d{1,3}\.?){4})$/', $ip, $matches ) ){
            $ipv4 = dechex( sprintf( "%u", ip2long( $matches[2] ) ) );
            $ip = $matches[1]
                  . (strlen($ipv4) > 4 ? substr( $ipv4, 0, -4) : '0:')
                  . ':' . substr($ipv4, -4, 4);
            
            // Normalizing hextets number (
        }elseif(strpos($ip, '::') !== false){
            $ip = str_replace('::', str_repeat(':0', 8 - substr_count($ip, ':')) . ':', $ip);
        }
        
        // Replacing head and rear ":" with "0:" and ":0"
        $ip = strpos($ip, ':') === 0         ? '0' . $ip : $ip;
        $ip = strpos(strrev($ip), ':') === 0 ? $ip . '0' : $ip;
        
        // Simplifying hextets. Replacing heading zeros in the each hextets
        if(preg_match('/:0(?=[a-z0-9]+)/', $ip)){
            $ip = preg_replace('/:0(?=[a-z0-9]+)/', ':', strtolower($ip));
            $ip = self::v6_normalize($ip);
        }

        return strtolower( $ip );
    }
    
    /**
     * Checks if the IP is in private range
     *
     * @param string $ip
     * @param string $ip_type
     *
     * @return bool
     */
    public static function is_private_network($ip, $ip_type = 'v4')
    {
        return self::mask_match($ip, self::$private_networks[$ip_type], $ip_type);
    }
    
    /**
     * Check if the IP belong to mask.  Recursive.
     * Octet by octet for IPv4
     * Hextet by hextet for IPv6
     *
     * @param string $ip
     * @param string $cidr       network to compare with
     * @param string $ip_type    IPv6 or IPv4
     * @param int    $xtet_count Recursive counter. Determs current part of address to check.
     *
     * @return bool
     */
    public static function mask_match($ip, $cidr, $ip_type = 'v4', $xtet_count = 0)
    {
        if(is_array($cidr)){
            foreach($cidr as $curr_mask){
                if(self::mask_match($ip, $curr_mask, $ip_type)){
                    return true;
                }
            }
            unset($curr_mask);
            return false;
        }
        
        if( ! self::validate( $ip ) || ! self::cidr__validate( $cidr ) ){
            return false;
        }
        
        $xtet_base = ($ip_type == 'v4') ? 8 : 16;
        
        // Calculate mask
        $exploded = explode('/', $cidr);
        $net_ip = $exploded[0];
        $mask = $exploded[1];
        
        // Exit condition
        $xtet_end = ceil($mask / $xtet_base);
        if($xtet_count == $xtet_end)
            return true;
        
        // Lenght of bits for comparsion
        $mask = $mask - $xtet_base * $xtet_count >= $xtet_base ? $xtet_base : $mask - $xtet_base * $xtet_count;
        
        // Explode by octets/hextets from IP and Net
        $net_ip_xtets = explode($ip_type == 'v4' ? '.' : ':', $net_ip);
        $ip_xtets = explode($ip_type == 'v4' ? '.' : ':', $ip);
        
        // Standartizing. Getting current octets/hextets. Adding leading zeros.
        $net_xtet = str_pad(decbin($ip_type == 'v4' ? $net_ip_xtets[$xtet_count] : @hexdec($net_ip_xtets[$xtet_count])), $xtet_base, 0, STR_PAD_LEFT);
        $ip_xtet = str_pad(decbin($ip_type == 'v4' ? $ip_xtets[$xtet_count] : @hexdec($ip_xtets[$xtet_count])), $xtet_base, 0, STR_PAD_LEFT);
        
        // Comparing bit by bit
        for($i = 0, $result = true; $mask != 0; $mask--, $i++){
            if($ip_xtet[$i] != $net_xtet[$i]){
                $result = false;
                break;
            }
        }
        
        // Recursing. Moving to next octet/hextet.
        if($result)
            $result = self::mask_match($ip, $cidr, $ip_type, $xtet_count + 1);
        
        return $result;
        
    }
    
    /**
     * Reduce IPv6
     *
     * @param string $ip
     *
     * @return string IPv6
     */
    public static function v6_reduce($ip)
    {
        if(strpos($ip, ':') !== false){
            $ip = preg_replace('/:0{1,4}/', ':', $ip);
            $ip = preg_replace('/:{2,}/', '::', $ip);
            $ip = strpos($ip, '0') === 0 ? substr($ip, 1) : $ip;
        }
        return $ip;
    }
    
    /**
     * Validate CIDR
     *
     * @param string $cidr expects string like 1.1.1.1/32
     *
     * @return bool
     */
    public static function cidr__validate( $cidr ){
        $cidr = explode( '/', $cidr );
        return isset( $cidr[0], $cidr[1] ) && self::validate( $cidr[0] ) && preg_match( '@\d{1,2}@', $cidr[1] );
    }
}